import unittest

from decentnet.modules.comm.beam import Beam


class TestBeamHash(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # This method is called before tests in an individual class are run
        pass

    def test_create_beam_hash_length(self):
        # Test that the hash length is correct for blake2b (default is 64 characters long)
        beam_hash = Beam.create_beam_hash(21112)
        self.assertEqual(len(beam_hash), 128,
                         "The hash should be 128 characters long for the default blake2b.")

    def test_create_beam_hash_is_hex(self):
        # Test that the hash consists only of hexadecimal characters
        beam_hash = Beam.create_beam_hash(21112)
        hex_digits = set("0123456789abcdef")
        # Convert hash to a lower case to simplify checking
        self.assertTrue(all(c in hex_digits for c in beam_hash.lower()),
                        "The hash should be a hexadecimal string.")


if __name__ == '__main__':
    unittest.main()



import math
import random
import string
import timeit
import unittest

from decentnet.modules.compression.wrapper import CompressionWrapper


def generate_random_data(size):
    # Generate random data with the specified size
    return bytes(''.join(random.choices(string.ascii_letters + string.digits, k=size)),
                 'utf-8')


def format_size(size_bytes):
    # Adapted from: https://stackoverflow.com/a/1094933/9667816
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    size = round(size_bytes / p, 2)
    return f"{size} {size_name[i]}"


class CompressionPerformanceTests(unittest.TestCase):
    def setUp(self) -> None:
        self.data = generate_random_data(
            random.randint(1000, 10000))

    def test_slow_compression_performance(self):
        def run_slow_compression():
            compressed_data = CompressionWrapper.compress_brotli(self.data, quality=10)
            original_size = len(self.data)
            compressed_size = len(compressed_data)
            saved_bytes = original_size - compressed_size
            print("Slow Compression: Compressed", format_size(original_size), "to",
                  format_size(compressed_size) + ". Saved",
                  format_size(saved_bytes) + ".")

        slow_time = timeit.timeit(run_slow_compression, number=1)

        print("Slow Compression Time:", slow_time * 1000, "milliseconds")

    def test_fast_compression_performance(self):
        def run_fast_compression():
            compressed_data = CompressionWrapper.compress_zlib(self.data, level=9)
            original_size = len(self.data)
            compressed_size = len(compressed_data)
            saved_bytes = original_size - compressed_size
            print("Fast Compression: Compressed", format_size(original_size), "to",
                  format_size(compressed_size) + ". Saved",
                  format_size(saved_bytes) + ".")

        fast_time = timeit.timeit(run_fast_compression, number=1)

        print("Fast Compression Time:", fast_time * 1000, "milliseconds")


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import os
from contextlib import asynccontextmanager
from functools import lru_cache
from pathlib import Path

from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import DB_FILENAME, DATABASE_URL
from decentnet.modules.db.models import Base
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


@lru_cache()
def get_root_dir() -> Path:
    return Path(os.path.abspath(__file__)).parent.parent.parent.parent


async def init_db(eng):
    async with eng.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)


db_file = get_root_dir() / DB_FILENAME
engine = create_async_engine(DATABASE_URL, echo=False)

asyncio.run(init_db(engine))


@asynccontextmanager
async def session_scope():
    """Provide a transactional scope around a series of operations asynchronously."""
    async_session = async_sessionmaker(
        bind=engine,
        expire_on_commit=False
    )()
    try:
        yield async_session
        await async_session.commit()
    except Exception as e:
        await async_session.rollback()
        raise e
    finally:
        await async_session.close()



import numpy as np
import tensorflow.compat.v2 as tf
from keras.layers import Dense, Dropout, Conv1D

tf.enable_v2_behavior()


def posterior_mean_field(kernel_size, bias_size=0.1, dtype=None):
    n = kernel_size + bias_size
    c = np.log(np.expm1(1.))
    return tf.keras.Sequential([
        tfp.layers.VariableLayer(2 * n, dtype=dtype),
        tfp.layers.DistributionLambda(lambda t: tfd.Independent(
            tfd.Normal(loc=t[..., :n],
                       scale=1e-5 + tf.nn.softplus(c + t[..., n:])),
            reinterpreted_batch_ndims=1)),
    ])


def prior_trainable(kernel_size, bias_size=0.01, dtype=None):
    n = kernel_size + bias_size
    return tf.keras.Sequential([
        tfp.layers.VariableLayer(n, dtype=dtype),
        tfp.layers.DistributionLambda(lambda t: tfd.Independent(
            tfd.Normal(loc=t, scale=1),
            reinterpreted_batch_ndims=1)),
    ])


import tensorflow_probability as tfp

tfd = tfp.distributions
samples = 2000

train = np.loadtxt("samples.txt", delimiter=",")

x_train = train[:, :5]
y_train = train[:, 5:]

epochs = round(samples + 0.15 * samples)

n_layer = (samples / 100) * x_train.shape[0]

model = tf.keras.Sequential([
    tfp.layers.DenseVariational(n_layer, posterior_mean_field, prior_trainable,
                                kl_weight=x_train / n_layer / x_train.shape[0]),
    tfp.layers.DistributionLambda(
        lambda t: tfd.Normal(loc=t[..., :1],
                             scale=1e-3 + tf.math.softplus(0.01 * t[..., 1:]))),
    Dense(8, activation="gelu"),
    Dense(1, activation="gelu")
])
model.compile(loss='mean_absolute_error',
              optimizer=tf.keras.optimizers.Adam(0.01))

model.fit(x_train, y_train, epochs=epochs)
predict = model.predict(np.array([[5, 15, 9, 9, 4]]))
print(predict)

points = 100000

take_profit = np.linspace(0.5, 25, num=points)
stop_loss = np.linspace(0.01, 30, num=points)
trail_stop_loss = np.linspace(0.01, 20, num=points)
trail_stop_arm_at = np.linspace(0.01, 20, num=points)
trail_buy = np.linspace(0.01, 10, num=points)

predict_stack = np.vstack([take_profit, stop_loss, trail_stop_loss,
                           trail_stop_arm_at, trail_buy]).transpose()

m = model.predict(predict_stack)
solved = np.concatenate((predict_stack, m), axis=1)

np.savetxt("points.txt", solved, delimiter=",")



import logging

from rich.logging import RichHandler

from decentnet.consensus.dev_constants import LOG_LEVEL


def setup_logger(debug: bool, logger):
    if not debug:
        logging.disable(logging.DEBUG)
    # root = logging.getLogger()
    # root.setLevel(LOG_LEVEL)
    logger.setLevel(LOG_LEVEL)
    logging.basicConfig(
        level=LOG_LEVEL if debug else logging.CRITICAL,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler()]
    )



import logging
import struct
import time
from typing import Union

from Crypto.Util.Padding import pad, unpad

from decentnet.consensus.block_sizing import MERGED_DIFFICULTY_BYTE_LEN, HASH_LEN, \
    INDEX_SIZE, NONCE_SIZE, TIMESTAMP_SIZE, HASH_LEN_BLOCK
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG, BLOCK_ERROR_DATA_LOG_LEN
from decentnet.modules.compression.wrapper import CompressionWrapper
from decentnet.modules.hash_type.hash_type import MemoryHash
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.difficulty import Difficulty
from decentnet.modules.pow.pow import PoW
from decentnet.modules.timer.timer import Timer

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Block:
    index: int
    previous_hash: bytes = None
    diff: Difficulty
    data: bytearray
    timestamp: float
    nonce: int | None
    _hash: MemoryHash | None
    ttc: float
    signature: str | None = None

    def __init__(self, index: int, prev_hash: bytes,
                 difficulty: Difficulty,
                 data: Union[bytearray, bytes], logs_dict: dict = None):
        self.index = index
        self.previous_hash = prev_hash
        self.data = data
        self.diff = difficulty
        self.timestamp = time.time()
        self.nonce = 0
        self.signature = None
        self.logs_dict = logs_dict

    def __str__(self):
        dd = str(bytes(self.data))
        display_data = dd[:BLOCK_ERROR_DATA_LOG_LEN] + "..." if len(
            dd) > BLOCK_ERROR_DATA_LOG_LEN else dd
        result = f"Block #{self.index}\n" \
                 f"Previous Hash: {self.previous_hash.hex()[2:].zfill(HASH_LEN) if self.index != 0 else 'GENESIS BLOCK'}\n" \
                 f"Difficulty: {self.diff}\n" \
                 f"Data: {display_data}\n" \
                 f"Timestamp: {self.timestamp}\n" \
                 f"Nonce: {self.nonce}\n"

        if hasattr(self, "_hash"):
            result += f"Hash: {self.hash.value.hex()[2:].zfill(HASH_LEN)}\n"

        return result

    @property
    def hash(self):
        return self._hash

    def compute_hash(self) -> MemoryHash:
        index_bytes = self.index.to_bytes(INDEX_SIZE, byteorder=ENDIAN_TYPE, signed=False)
        diff_bytes = self.diff.to_bytes()
        previous_hash_bytes = self.previous_hash
        timestamp_bytes = struct.pack('d', self.timestamp)

        packed_block = (
                index_bytes + diff_bytes + previous_hash_bytes + timestamp_bytes + self.data)

        self._hash = MemoryHash(self.diff, packed_block)
        return self._hash

    def to_bytes(self) -> bytes:
        index_bytes = self.index.to_bytes(INDEX_SIZE, byteorder=ENDIAN_TYPE, signed=False)
        # Difficulty bytes (assuming it's already in bytes form, otherwise adapt accordingly)
        diff_bytes = self.diff.to_bytes()  # Replace with actual conversion if necessary

        # Pad the previous hash and convert it to bytes
        previous_hash_bytes = pad(bytes(self.previous_hash), HASH_LEN_BLOCK, style='pkcs7')

        nonce_bytes = self.nonce.to_bytes(NONCE_SIZE, byteorder=ENDIAN_TYPE,
                                          signed=False)
        timestamp_bytes = struct.pack('d', self.timestamp)

        compressed_data = self.data if self.diff.compression_level == 0 else CompressionWrapper.compress_lz4(
            self.data, self.diff.compression_level)

        packed_block = (index_bytes + diff_bytes + previous_hash_bytes +
                        nonce_bytes + timestamp_bytes + compressed_data)  # Speed up compression

        packed_block_len = len(packed_block)

        if self.logs_dict:
            compressed_data_len = len(compressed_data)
            self.logs_dict["data_header_ratio"][1].send(compressed_data_len / packed_block_len)

        logger.debug(f"Packed Block into {packed_block_len} B")
        return packed_block

    @classmethod
    def from_bytes(cls, compressed_block_bytes: bytes):
        block = cls.__new__(cls)

        block_bytes = memoryview(compressed_block_bytes)

        cursor = 0

        # Unpack index
        block.index = int.from_bytes(block_bytes[cursor:cursor + INDEX_SIZE], byteorder=ENDIAN_TYPE,
                                     signed=False)
        cursor += INDEX_SIZE

        # Unpack difficulty
        block.diff = Difficulty.from_bytes(block_bytes[cursor:cursor + MERGED_DIFFICULTY_BYTE_LEN])
        cursor += MERGED_DIFFICULTY_BYTE_LEN

        # Unpack previous hash
        block.previous_hash = unpad(bytes(block_bytes[cursor:cursor + HASH_LEN_BLOCK]), HASH_LEN_BLOCK,
                                    style='pkcs7')
        cursor += HASH_LEN_BLOCK

        # Unpack nonce
        block.nonce = int.from_bytes(block_bytes[cursor:cursor + NONCE_SIZE],
                                     byteorder=ENDIAN_TYPE,
                                     signed=False)
        cursor += NONCE_SIZE

        # Unpack timestamp
        block.timestamp = struct.unpack("d", block_bytes[cursor:cursor + TIMESTAMP_SIZE])[0]
        cursor += TIMESTAMP_SIZE

        # Decompress the remaining data
        block.data = block_bytes[
                     cursor:] if block.diff.compression_level == 0 else CompressionWrapper.decompress_lz4(
            block_bytes[cursor:])

        return block

    def mine(self, measure=False):
        logger.debug(f"Mining block #{self.index}")
        if measure:
            t = Timer()

        a = self.compute_hash()

        finished_hash, finished_nonce = PoW.compute(a, self.diff.n_bits)
        self.nonce = finished_nonce

        if measure:
            self.ttc = t.stop()
            return finished_nonce, self.ttc
        else:
            return finished_nonce



"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the py-package-template package.
"""

__title__ = "decentmesh"
__description__ = (
    "Simple connection over decentralized network"
)

__version__ = "0.0.54"
__author__ = "Jiri Otoupal"
__author_email__ = "jiri-otoupal@ips-database.eu"
__license__ = "JO-CAL"
__url__ = "https://github.com/jiri-otoupal/DecentNet-Py"
__pypi_repo__ = "https://pypi.org/project/decentmesh/"

___nv = __version__.split(".")
NETWORK_VERSION = bytes((int(___nv[0]), int(___nv[1]), int(___nv[2])))



import logging
import multiprocessing
from threading import Thread
from wsgiref.simple_server import make_server

import prometheus_client
from prometheus_client import Histogram

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import PROMETHEUS_PORT, PROMETHEUS_HOST
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

block_difficulty_bits = Histogram('block_difficulty_bits', 'Difficulty of blockchain bits')
block_difficulty_memory = Histogram('block_difficulty_memory',
                                    'Difficulty of blockchain memory')
block_difficulty_time = Histogram('block_difficulty_time', 'Difficulty of blockchain time')
block_difficulty_parallel = Histogram('block_difficulty_parallel',
                                      'Difficulty of blockchain parallel')
data_header_ratio = Histogram('data_header_ratio', 'Header/Data Ratio')
prom_block_process_time = Histogram('block_process_time', 'Time spent processing a block')
prom_data_received = Histogram("data_received_for_relay", "Data received for relaying")
prom_path_length = Histogram("path_length", "Path length for relaying data")
prom_data_published = Histogram("data_published_for_relay", "Data published for relaying")


class Stats:
    @staticmethod
    def start_prometheus_server(log_dict: dict):
        log_dict["block_difficulty_bits"] = multiprocessing.Pipe()
        log_dict["block_difficulty_memory"] = multiprocessing.Pipe()
        log_dict["block_difficulty_time"] = multiprocessing.Pipe()
        log_dict["block_difficulty_parallel"] = multiprocessing.Pipe()
        log_dict["data_header_ratio"] = multiprocessing.Pipe()
        log_dict["block_process_time"] = multiprocessing.Pipe()
        log_dict["data_received_for_relay"] = multiprocessing.Pipe()
        log_dict["path_length"] = multiprocessing.Pipe()
        log_dict["data_published_for_relay"] = multiprocessing.Pipe()
        log_dict["prom_path_length"] = multiprocessing.Pipe()
        log_dict["prom_data_received"] = multiprocessing.Pipe()
        log_dict["prom_data_published"] = multiprocessing.Pipe()
        log_dict["prom_block_process_time"] = multiprocessing.Pipe()

        Stats.wait_write_stats(log_dict)
        logger.debug("Starting prometheus client...")
        app = prometheus_client.make_wsgi_app()
        httpd = make_server(PROMETHEUS_HOST, PROMETHEUS_PORT, app)
        httpd.serve_forever()

        for value in log_dict.values():
            value[0].close()
            value[1].close()

    @staticmethod
    def wait_write_stats(log_dict: dict):
        for key, log_pipe in log_dict.items():
            Thread(target=Stats.wait_write, args=(key, log_pipe), name=f"Observing {key}",
                   daemon=True).start()

    @staticmethod
    def wait_write(key, log_pipe):
        while True:
            value = log_pipe[0].recv()
            globals().get(key).observe(value)



import asyncio
import logging
from collections import deque

import cbor2
import pyximport

from decentnet.__version__ import NETWORK_VERSION
from decentnet.consensus.blockchain_params import BlockchainParams, SAVE_BLOCKS_TO_DB_DEFAULT
from decentnet.consensus.cmd_enum import CMD
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.blockchain.block import Block
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import BlockchainTable, BlockTable, BlockSignatureTable
from decentnet.modules.key_util.key_manager import KeyManager
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.passgen.passgen import SecurePasswordGenerator
from decentnet.modules.pow.difficulty import Difficulty
from decentnet.modules.pow.pow_utils import PowUtils
from decentnet.modules.timer.timer import Timer

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

pyximport.install()
try:
    from decentnet.modules.convert.byte_to_base64_fast import bytes_to_base64  # noqa
except ImportError as ex:
    logger.debug(f"Operating in slow mode because of {ex}")
    logger.warning("Base64 convert module is operating in slow mode")
    from decentnet.modules.convert.byte_to_base64_slow import bytes_to_base64


class Blockchain:
    _id: int
    chain: deque[Block]
    version: bytes
    _difficulty: Difficulty

    def __init__(self, genesis_data: str | None = None, save=SAVE_BLOCKS_TO_DB_DEFAULT,
                 difficulty: Difficulty = BlockchainParams.seed_difficulty,
                 pub_key_for_encryption: str | None = None,
                 beam_id: int | None = None,
                 next_difficulty: Difficulty = BlockchainParams.low_diff, name="default",
                 logs_dict: dict = None):
        self.version = NETWORK_VERSION
        self.pub_key_for_encryption = pub_key_for_encryption
        self._difficulty = difficulty
        self.chain: deque[Block] = deque()
        self.save = save
        self.name = name
        self._id = -1
        self.logs_dict = logs_dict

        if save:
            asyncio.run(self.__save(beam_id))

        if genesis_data:
            if self.pub_key_for_encryption is None:
                logger.warning(
                    "Public key for encryption not provided, data will be not encrypted.")

            seed_block = self.create_genesis(genesis_data.encode("ascii"),
                                             owned_public_key_for_encryption=self.pub_key_for_encryption,
                                             difficulty=difficulty,
                                             next_difficulty=next_difficulty, logs_dict=logs_dict)
            seed_block.mine()
            if not self.insert(seed_block):
                raise Exception("Failed to insert genesis block")

            self.difficulty = next_difficulty

    async def __save(self, beam_id):
        async with session_scope() as session:
            b_gen = BlockchainTable(
                version=self.version.hex(),
                beam_id=beam_id,
                difficulty=str(self.difficulty)
            )
            session.add(b_gen)
            await session.commit()
            self._id = b_gen.blockchain_id

    def __len__(self):
        return len(self.chain)

    @classmethod
    def create_genesis(cls, data, owned_public_key_for_encryption: str | None,
                       difficulty: Difficulty = BlockchainParams.seed_difficulty,
                       next_difficulty: Difficulty = BlockchainParams.low_diff, logs_dict: dict = None):

        genesis_data = cbor2.dumps({"data": data.decode("ascii"),
                                    "new_diff": next_difficulty.to_bytes().decode("ascii"),
                                    "enc_pub_key": owned_public_key_for_encryption})
        block = Block(0, b"0", difficulty, genesis_data, logs_dict)
        block.seed_hash = block.compute_hash()
        return block

    @classmethod
    def convert_handshake_block_dict_to_bytes(cls, handshake_block: dict) -> bytes:
        return cbor2.dumps(handshake_block)

    @classmethod
    def create_handshake_encryption_block_raw(cls,
                                              public_key_received_for_encryption: str,
                                              bits: int = 192, algorithm: str = "AES",
                                              additional_data: str = "") -> (bytes, bytes):
        """
        Create handshake encryption data for block
        :param additional_data:
        :param public_key_received_for_encryption:
        :param bits:
        :param algorithm:
        :return: handshake block in raw byte format dumped by cbor2, password
        """
        logger.debug(f"Creating handshake block with pub_enc {public_key_received_for_encryption}")
        encrypted_password, password = cls._create_handshake_block_base(bits,
                                                                        public_key_received_for_encryption)

        data = cls.convert_handshake_block_dict_to_bytes(
            {"cmd": CMD.HANDSHAKE_ENCRYPTION.value, "key": encrypted_password,
             "data": additional_data,
             "algo": algorithm, "bits": bits})
        return data, password

    @classmethod
    def _create_handshake_block_base(cls, bits: int, public_key_received_for_encryption: str):
        """
        Creates base data for handshake block
        :param bits:
        :param public_key_received_for_encryption:
        :return:
        """
        byte_length = int(bits / 8)
        password = SecurePasswordGenerator(length=byte_length).generate().encode("utf-8")[:byte_length]
        pub_key = AsymCrypt.public_key_from_base64(public_key_received_for_encryption,
                                                   can_encrypt=True)
        encrypted_password_bytes = AsymCrypt.encrypt_message(pub_key,
                                                             password)
        encrypted_password = bytes_to_base64(encrypted_password_bytes)
        return encrypted_password, password

    @classmethod
    def create_handshake_encryption_block_dict(cls,
                                               public_key_received_for_encryption: str,
                                               bits: int = 192, algorithm: str = "AES",
                                               additional_data: str = "") -> (dict, bytes):
        """
        Create handshake encryption data for block
        :param additional_data:
        :param public_key_received_for_encryption:
        :param bits:
        :param algorithm:
        :return: handshake block in dict format, password
        """
        encrypted_password, password = cls._create_handshake_block_base(bits,
                                                                        public_key_received_for_encryption)

        data = {"cmd": CMD.HANDSHAKE_ENCRYPTION.value,
                "key": encrypted_password,
                "data": additional_data,
                "algo": algorithm, "bits": bits}
        return data, password

    def get_last(self):
        return self.chain[-1]

    @staticmethod
    def get_next_work_required(requested_diff: Difficulty) -> int:
        """

        Get Required work from Difficulty parameters
        @param requested_diff Input Difficulty
        @return

        :param requested_diff:
        :return:
        """
        return 1 << requested_diff.n_bits

    def validate_next_block(self, block: Block):

        if block.index != 0 and block.index != self.get_last().index + 1:
            logger.error(
                f"Block index is not same as chain length {block.index} != {len(self.chain)}")
            return False

        if block.index > 0 and block.previous_hash != (cb := self.get_last().hash.value):
            logger.error(
                f"Previous hash is not correct {block.previous_hash.hex()} != {cb.value.hex()}")
            return False

        # Check if the hash satisfies the difficulty requirements
        block_hash = block.compute_hash()
        block_z_bits = PowUtils.get_bit_length(block_hash, block.nonce, block.diff)
        target_bits = self.difficulty.hash_len_chars * 8 - self.difficulty.n_bits

        if block_z_bits > target_bits:
            logger.debug("Block hash %s %s nonce %s" % (block_hash.value_as_hex(), block_z_bits, block.nonce))
            logger.error(
                f"Mined block was not mined with correct difficulty, or not mined properly {block_z_bits} b > {target_bits}")
            return False

        if block.diff != self.difficulty:
            logger.error(
                f"Block difficulty is not same, Block: {block.diff} != Blockchain: {self.difficulty}")
            return False

        # Check if the timestamp is greater than the last block's timestamp
        # IDK if this check should be here
        if block.index > 0 and block.timestamp < self.get_last().timestamp:
            logger.error(
                f"Timestamp is traveling back in time {block.timestamp} < {self.get_last().timestamp}")
            return False

        return True

    def clear(self):
        self.chain.clear()

    def insert(self, block: Block):
        if RUN_IN_DEBUG:
            t = Timer()

        if self.validate_next_block(block):
            if RUN_IN_DEBUG:
                logger.debug("Validation took %s ms" % (t.stop()))
            self.chain.append(block)

            if RUN_IN_DEBUG:
                logger.debug("Append took %s ms" % (t.stop()))

            if self.save:
                asyncio.run(self._save_block(block))
            if RUN_IN_DEBUG:
                logger.debug("Inserted! %s into %s" % (block, self.name))
            return True
        logger.error(f"Block {block} validation failed")
        logger.error(f"Blockchain that has failed to insert block \n{self}")
        return False

    def __str__(self):
        out = f"Name: {self.name} | ID: {self.id} | Blockchain Version: {self.version}  | Blocks:\n"
        for block in self.chain:
            out += f"{block}"
        return out

    def template_next_block(self, requested_diff: Difficulty, data: bytes | bytearray):
        """
        Template a new block without setting a new difficulty
        :param requested_diff:
        :param data:
        :return:
        """
        logger.debug("Templated next block with difficulty %s" % requested_diff)
        last = self.get_last()
        return Block(last.index + 1, last.hash.value, requested_diff, data, self.logs_dict)

    @property
    def difficulty(self):
        return self._difficulty

    @difficulty.setter
    def difficulty(self, value: Difficulty):
        self._difficulty = value
        logger.debug(f"Set new blockchain difficulty {value}")
        if self.logs_dict:
            self.logs_dict["block_difficulty_bits"][1].send(value.n_bits)
            self.logs_dict["block_difficulty_time"][1].send(value.t_cost)
            self.logs_dict["block_difficulty_memory"][1].send(value.m_cost)
            self.logs_dict["block_difficulty_parallel"][1].send(value.p_cost)

    async def _save_block(self, block):
        async with session_scope() as session:
            # Create a new BlockTable entry
            bdb = BlockTable(
                blockchain_id=self._id,
                index=block.index,
                hash=bytearray(block.hash.value),
                previous_hash=block.previous_hash,
                data=block.data,
                nonce=block.nonce
            )
            session.add(bdb)

            # Flush the session asynchronously to persist the block and get its ID
            await session.flush()

            # If the block has a signature, save it in the BlockSignatureTable
            if block.signature:
                bst = BlockSignatureTable(
                    block_id=bdb.block_id,
                    signature=KeyManager.key_to_base64(block.signature)
                )
                session.add(bst)

            # Commit the transaction asynchronously
            await session.commit()

    @property
    def id(self):
        return self._id
